# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of BOlib.
#
#    BOlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    BOlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with BOlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np


class OptimizationMethod(object):
    """

    """
    def __init__(self):
        """

        """

        self.log = dict()

    def set_seed(self, seed):
        """

        :param seed:
        :type seed:
        :return:
        :rtype:
        """
        self.log['seed'] = seed
        np.random.seed(seed)

    def random_sample(self, bounds, batch_size=1):
        """

        :param bounds:
        :type bounds:
        :param batch_size:
        :type batch_size:
        :return:
        :rtype:
        """
        x_t = np.vstack([
            np.random.uniform(
                low=lower,
                high=upper,
                size=batch_size
            )
            for lower, upper in bounds
        ]).T

        return x_t

    def minimize(self, fun, x0, args=(), **kwargs):
        """

        :param fun:
        :type fun:
        :param x0:
        :type x0:
        :param args:
        :type args:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        raise NotImplementedError("Not Implemented. This is an interface.")



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of BOlib.
#
#    BOlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    BOlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with BOlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from .objective_function import ObjectiveFunction


class Branin(ObjectiveFunction):
    """

    """
    def __init__(self):
        """

        """
        super(Branin, self).__init__(
            d=2,
            gaussian_noise=0,
            f_bias=0.0,
            max_eval=200,
            lower=[-5, 0],
            upper=[10, 15],
            objective=[9.42478, 2.475],
            objective_val=0.397887,
            params=['x', 'y'],
            types=[float, float]
        )

    def batch_evaluate(self, points):
        """
        Branin test function

        The number of variables n = 2.
        constraints:
        -5 <= x <= 10, 0 <= y <= 15
        three global optima:  (-pi, 12.275), (pi, 2.275), (9.42478, 2.475),
        where branin = 0.397887

        :param points:
        :type points:
        :return:
        :rtype:
        """
        x = points[:, 0][:, None]
        y = points[:, 1][:, None]
        result = np.power((y-(5.1/(4 * np.power(np.pi, 2))) *
                           np.power(x, 2)+5 * x/np.pi-6), 2)
        result += 10*(1-1/(8*np.pi))*np.cos(x)+10
        return result + self.f_bias



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of BOlib.
#
#    BOlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    BOlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with BOlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from .objective_function import ObjectiveFunction


class Hartmann(ObjectiveFunction):
    """

    """
    def __init__(self):
        """

        """
        super(Hartmann, self).__init__(
            d=6,
            gaussian_noise=0,
            f_bias=0.0,
            max_eval=200,
            lower=[0, 0, 0, 0, 0, 0],
            upper=[1, 1, 1, 1, 1, 1],
            objective=[0.20169, 0.150011, 0.476874,
                       0.275332, 0.311652, 0.6573],
            objective_val=-3.32236,
            params=['x', 'y', 'z', 'a', 'b', 'c'],
            types=[float, float, float, float, float, float]
        )

    def batch_evaluate(self, points):
        """
        6d Hartmann test function

        constraints:
        0 <= xi <= 1, i = 1..6
        global optimum at (0.20169, 0.150011, 0.476874,
                           0.275332, 0.311652, 0.6573)
        where har6 = -3.32236

        :param points:
        :type points:
        :return:
        :rtype:
        """
        x = points[:, 0][:, None]
        y = points[:, 1][:, None]
        z = points[:, 2][:, None]
        a = points[:, 3][:, None]
        b = points[:, 4][:, None]
        c = points[:, 5][:, None]
        value = np.array([x, y, z, a, b, c])

        a = np.array([[10.0, 3.0, 17.0, 3.5, 1.7, 8.0],
                      [0.05, 10.0, 17.0, 0.1, 8.0, 14.0],
                      [3.0, 3.5, 1.7, 10.0, 17.0, 8.0],
                      [17.0, 8.0, 0.05, 10.0, 0.1, 14.0]])
        c = np.array([1.0, 1.2, 3.0, 3.2])
        p = np.array([[0.1312, 0.1696, 0.5569, 0.0124, 0.8283, 0.5886],
                      [0.2329, 0.4135, 0.8307, 0.3736, 0.1004, 0.9991],
                      [0.2348, 0.1451, 0.3522, 0.2883, 0.3047, 0.6650],
                      [0.4047, 0.8828, 0.8732, 0.5743, 0.1091, 0.0381]])
        s = 0
        for i in [0, 1, 2, 3]:
            sm = a[i, 0]*(value[0]-p[i, 0])**2
            sm += a[i, 1]*(value[1]-p[i, 1])**2
            sm += a[i, 2]*(value[2]-p[i, 2])**2
            sm += a[i, 3]*(value[3]-p[i, 3])**2
            sm += a[i, 4]*(value[4]-p[i, 4])**2
            sm += a[i, 5]*(value[5]-p[i, 5])**2
            s += c[i]*np.exp(-sm)
        result = s
        return - result + self.f_bias



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of BOlib.
#
#    BOlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    BOlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with BOlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from .objective_function import ObjectiveFunction


class Camelback(ObjectiveFunction):
    """

    """
    def __init__(self):
        """

        """
        super(Camelback, self).__init__(
            d=2,
            gaussian_noise=0,
            f_bias=0.0,
            max_eval=100,
            lower=[-2, -1],
            upper=[2, 1],
            objective=[0.0898, -0.7126],
            objective_val=-1.0316,
            params=['x', 'y'],
            types=[float, float]
        )

    def batch_evaluate(self, points):
        """
        Camelback test function

        The number of variables n = 2.
        constraints:
        -2 <= x <= 2, -1 <= y <= 1
        two global minima at (0.0898, -0.7126) and (-0.0898, 0.7126) where
        f(x) = -1.0316

        :param points:
        :type points:
        :return:
        :rtype:
        """
        x = points[:, 0][:, None]
        y = points[:, 1][:, None]
        x2 = np.power(x, 2)
        x4 = np.power(x, 4)
        y2 = np.power(y, 2)
        result = (np.multiply((4 - 2.1*x2 + x4/3), x2) +
                  np.multiply(x, y) +
                  np.multiply((-4+4*y2), y2))
        return result + self.f_bias



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of BOlib.
#
#    BOlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    BOlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with BOlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import scipy.stats as sps

from .acquisition_function import AcquisitionFunction


class ProbabilityOfImprovement(AcquisitionFunction):
    """

    """
    def __init__(self):
        """

        """
        self.xi = 0.00

        super(ProbabilityOfImprovement, self).__init__()

    def batch_evaluate(self, points, dk_dx_needed=False):
        """
        Probability of Improvement Acquisition Function Implementation

        :param points:
        :type points:
        :param dk_dx_needed:
        :type dk_dx_needed:
        :return:
        :rtype:
        """
        y_mean = self.model.mean_function.marginalize_mean(
            self.data['X'])

        mean = self.model.mean_function.marginalize_mean(
            points)

        var = self.model.covariance_function.marginalize_covariance(
            points, only_diagonal=True)

        sdev = np.sqrt(var)

        F = np.min(y_mean) - mean - self.xi
        sdev[sdev == 0.0] = np.inf
        Z = F / sdev
        result = sps.norm.cdf(Z)

        if not dk_dx_needed:
            return -result

        # TODO grad needed
        raise Exception('gradient not implemented yet')



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of BOlib.
#
#    BOlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    BOlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with BOlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np

class ObjectiveFunction(object):
    """

    """
    def __init__(self, d, gaussian_noise, f_bias, max_eval,
                 lower, upper, objective, objective_val, params, types):
        """

        """
        self.d = d
        self.gaussian_noise = gaussian_noise
        self.f_bias = f_bias
        self.max_eval = max_eval
        self.lower = lower
        self.upper = upper
        self.objective = objective
        self.objective_val = objective_val
        self.params = params
        self.type = types

    def evaluate(self, point, add_noise=True):
        """

        :return:
        :rtype:
        """
        point = np.array(point)

        if point.ndim != 1 and point.ndim != 2:
            raise Exception("Too many dimensions")

        points = point
        if point.ndim == 1:
            points = point[None, :]

        result = self.batch_evaluate(points)

        if point.ndim == 1:
            result = result[0, 0]

        if add_noise:
            result += np.random.randn(*result.shape) * self.gaussian_noise

        return result

    def batch_evaluate(self, points):
        """

        :return:
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    def get_objective(self):
        """

        :return:
        :rtype:
        """

        return self.objective

    def get_objective_val(self):
        """

        :return:
        :rtype:
        """

        return self.objective_val

    def get_max_eval(self):
        """

        :return:
        :rtype:
        """

        return self.max_eval

    def get_bounds(self):
        """

        :return:
        :rtype:
        """

        return [
            (self.lower[i], self.upper[i])
            for i in range(self.d)
        ]



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of BOlib.
#
#    BOlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    BOlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with BOlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from .objective_function import ObjectiveFunction


class Schwefel(ObjectiveFunction):
    """

    """
    def __init__(self):
        """

        """
        super(Schwefel, self).__init__(
            d=2,
            gaussian_noise=0,
            f_bias=-450,
            max_eval=400,
            lower=[-100, -100],
            upper=[100, 100],
            objective=[1.42478, 5.475],
            objective_val=-450,
            params=['x', 'y'],
            types=[float, float]
        )

    def batch_evaluate(self, points):
        """
        CEC 2005 F2 Unimodal Function

        :param points:
        :type points:
        :return:
        :rtype:
        """
        norm_points = points - self.objective

        result = 0
        for i in range(0, norm_points.shape[1]):
            element = 0
            for j in range(0, i+1):
                element += norm_points[:, j][:, None]
            result += np.power(element, 2)
        return result + self.f_bias



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of BOlib.
#
#    BOlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    BOlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with BOlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import scipy.optimize as spo

import gplib

from .random_grid import RandomGrid
from .sequential_optimization import SequentialOptimization


class BayesianOptimization(SequentialOptimization):
    """

    """
    def __init__(self, model, fitting_method, validation, af):
        """

        """

        self.orig_model = model
        self.fitting_method = fitting_method
        self.validation = validation
        self.af = af

        super(BayesianOptimization, self).__init__()

    def next_sample(self, data, bounds):
        """

        :param data:
        :type data:
        :param bounds:
        :type bounds:
        :return:
        :rtype:
        """
        transformation = gplib.dm.DataTransformation(
            {
                'X': np.array(bounds, dtype=np.float).T,
                'Y': data['Y']
            }
        )
        t_data = transformation.transform(data)
        t_bounds = [(0., 1.) for _ in range(len(bounds))]

        fit_log = self.fitting_method.fit(
            model=self.orig_model,
            folds=self.validation.get_folds(t_data),
        )
        model = self.orig_model.get_posterior(t_data)

        self.af.set_model(model)
        self.af.set_data(t_data)

        # TODO Use better optimization methods
        optimization_method = RandomGrid()

        # Send next sample to Evaluate
        result = spo.minimize(
            self.af.evaluate,
            None,
            bounds=t_bounds,
            method=optimization_method.minimize,
            options={
                'maxiter': 20000
            }
        )

        x_sel = transformation.transform_key(result.x, 'X', inv=True)
        af_sel = result.fun

        return x_sel, {
            'x_sel': x_sel,
            'af_sel': af_sel,
            'fit_log': fit_log
        }



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of BOlib.
#
#    BOlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    BOlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with BOlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import scipy.optimize as spo

from .optimization_method import OptimizationMethod


class SequentialOptimization(OptimizationMethod):
    """

    """

    def minimize(self, fun, x0, args=(),
                 bounds=(), tol=1e-06, maxiter=200, disp=True,
                 **unknown_options):
        """

        :param fun:
        :type fun:
        :param x0:
        :type x0:
        :param args:
        :type args:
        :param bounds:
        :type bounds:
        :param tol:
        :type tol:
        :param maxiter:
        :type maxiter:
        :param disp:
        :type disp:
        :param unknown_options:
        :type unknown_options:
        :return:
        :rtype:
        """

        self.log['step_log'] = []

        data = {
            'X': np.empty((0, len(bounds))),
            'Y': np.empty((0, 1))
        }

        solver_log = {}
        x_t = x0

        improvement = 0.0
        funcalls = 0
        iterations = 0

        while funcalls < maxiter and not 0.0 < improvement < tol:
            # Sample objetive function
            x_t = x_t[:(maxiter-funcalls), :]
            y_t = fun(x_t)

            # Augment the data
            data['X'] = np.vstack((data['X'], x_t))
            data['Y'] = np.vstack((data['Y'], y_t))

            # Log best and current
            improvement = 0.0
            if len(y_t) != len(data['Y']):
                improvement = np.min(data['Y'][:-len(y_t), :]) - \
                              np.min(y_t)
            funcalls += x_t.shape[0]
            iterations += 1

            # Log best and current
            step_log = {
                'funcalls': funcalls,
                'iterations': iterations,
                'improvement': improvement,
                'x_t': x_t,
                'y_t': y_t,
                'x_best': data['X'][np.argmin(data['Y']), :],
                'f_best': np.min(data['Y']),
                'solver_log': solver_log
            }

            self.log['step_log'].append(step_log)
            if disp:
                print(step_log)

            # Choose next point(s) to evaluate
            x_t, solver_log = self.next_sample(data, bounds)

        return spo.OptimizeResult(
            fun=self.log['step_log'][-1]['f_best'],
            x=self.log['step_log'][-1]['x_best'],
            nit=iterations,
            nfev=funcalls,
            success=(iterations > 1)
        )

    def next_sample(self, data, bounds):
        """

        :param data:
        :type data:
        :param bounds:
        :type bounds:
        :return:
        :rtype:
        """
        raise NotImplementedError("Not Implemented. This is an interface.")



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of BOlib.
#
#    BOlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    BOlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with BOlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import scipy.optimize as spo

from .optimization_method import OptimizationMethod


class RandomGrid(OptimizationMethod):
    """

    """

    def minimize(self, fun, x0, args=(),
                 bounds=(), maxiter=200,
                 x0_square_size=1e-05, x0_square_samples=0.1,
                 **unknown_options):
        """

        :param fun:
        :type fun:
        :param x0:
        :type x0:
        :param args:
        :type args:
        :param bounds:
        :type bounds:
        :param maxiter:
        :type maxiter:
        :param x0_square_size:
        :type x0_square_size:
        :param x0_square_samples:
        :type x0_square_samples:
        :param unknown_options:
        :type unknown_options:
        :return:
        :rtype:
        """

        # Sample more points around x0
        x_t1 = self.random_sample(
            bounds,
            int(maxiter * (1.0-x0_square_samples))
        )

        y_t1 = fun(x_t1)

        x0 = x_t1[np.argmin(y_t1), :]

        x_t2 = self.random_sample(
            [
                (
                    x0_i - (x0_square_size * (bounds_i[1]-bounds_i[0])),
                    x0_i + (x0_square_size * (bounds_i[1]-bounds_i[0]))
                )
                for x0_i, bounds_i in zip(x0, bounds)
            ],
            int(maxiter * x0_square_samples)
        )

        y_t2 = fun(x_t2)

        x_t = np.vstack((x_t1, x_t2))
        y_t = np.vstack((y_t1, y_t2))

        best = np.argmin(y_t)

        x_best = x_t[best, :][None, :]
        y_best = y_t[best, 0]

        return spo.OptimizeResult(
            fun=y_best,
            x=x_best,
            nit=1,
            nfev=maxiter,
            success=True
        )



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of BOlib.
#
#    BOlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    BOlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with BOlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np


class AcquisitionFunction(object):
    """

    """
    def __init__(self):
        """

        """
        self.model = None
        self.data = None

    def set_model(self, model):
        """

        :param model:
        :type model:
        :return:
        :rtype:
        """
        self.model = model

    def set_data(self, data):
        """

        :param data:
        :type data:
        :return:
        :rtype:
        """
        self.data = data

    def evaluate(self, point, dk_dx_needed=False):
        """

        :param point:
        :type point:
        :param dk_dx_needed:
        :type dk_dx_needed:
        :return:
        :rtype:
        """
        point = np.array(point)

        if point.ndim != 1 and point.ndim != 2:
            raise Exception("Too many dimensions")

        points = point
        if point.ndim == 1:
            points = point[None, :]

        result = self.batch_evaluate(points, dk_dx_needed)

        if point.ndim == 1:
            result = result[0, 0]

        return result

    def batch_evaluate(self, points, dk_dx_needed=False):
        """

        :param points:
        :type points:
        :param dk_dx_needed:
        :type dk_dx_needed:
        :return:
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")



# TODO of_gp
# TODO svm_on_grid
# TODO lda_on_grid
from .branin import Branin
from .camelback import Camelback
from .hartmann import Hartmann
from .rastrigin import Rastrigin
from .rosenbrock import Rosenbrock
from .schwefel import Schwefel
from .sphere import Sphere



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of BOlib.
#
#    BOlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    BOlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with BOlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from .objective_function import ObjectiveFunction


class Rastrigin(ObjectiveFunction):
    """

    """
    def __init__(self):
        """

        """
        super(Rastrigin, self).__init__(
            d=2,
            gaussian_noise=0,
            f_bias=-330.0,
            max_eval=400,
            lower=[-5, -5],
            upper=[5, 5],
            objective=[1.42478, 0.475],
            objective_val=-330.0,
            params=['x', 'y'],
            types=[float, float]
        )

    def batch_evaluate(self, points):
        """
        CEC 2005 F9 Basic Multimodal Function

        :param points:
        :type points:
        :return:
        :rtype:
        """
        norm_points = points - self.objective

        result = 0.0
        for i in range(0, norm_points.shape[1]):
            result += np.power(norm_points[:, i][:, None], 2) - \
                10*np.cos(2*np.pi*norm_points[:, i][:, None])+10
        return result + self.f_bias
